// Copyright 2020, 2021 Francesco Biscani (bluescarni@gmail.com), Dario Izzo (dario.izzo@gmail.com)
//
// This file is part of the heyoka library.
//
// This Source Code Form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

#ifndef HEYOKA_CONFIG_HPP
#define HEYOKA_CONFIG_HPP

// clang-format off
#define HEYOKA_VERSION_STRING "0.17.0"
#define HEYOKA_VERSION_MAJOR 0
#define HEYOKA_VERSION_MINOR 17
#define HEYOKA_VERSION_PATCH 0
// clang-format on

/* #undef HEYOKA_WITH_MPPP */
#define HEYOKA_WITH_SLEEF
/* #undef HEYOKA_BUILD_STATIC_LIBRARY */

#if defined(HEYOKA_WITH_MPPP)

#include <mp++/config.hpp>

#if defined(MPPP_WITH_QUADMATH)

#define HEYOKA_HAVE_REAL128

#endif

#endif

#if defined(__i386) || defined(_M_IX86) || defined(_X86_) || defined(__amd64__) || defined(_M_AMD64)

#define HEYOKA_ARCH_X86

#elif defined(_ARCH_PPC) || defined(_M_PPC)

#define HEYOKA_ARCH_PPC

#endif

#endif
